!-------------------------------------------------------------------------------
!
!  Quasicontinuum (QC) Method: Mixed Continuum and Atomistic Simulation Package
!  QC Package distribution version 1.4 (November 2011)
!
!  Copyright (C) 2003 R. Miller, M. Ortiz, R. Phillips, D. Rodney, E. B. Tadmor
!  Copyright (C) 2004, 2005, 2006, 2007, 2011 R. Miller, E. B. Tadmor
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
!
!  For more information please visit the QC website at:
!
!    www.qcmethod.com
!
!  or contact Ron Miller or Ellad Tadmor with contact information below:
!
!  NOTE HOWEVER THAT NEITHER RON MILLER NOR ELLAD TADMOR WILL BE ABLE TO
!  PROVIDE SUPPORT OR DEBUGGING SERVICES FOR THE CODE. THE CODE IS SUPPLIED 
!  AS IS FOR YOUR USE.
!
!    Ron Miller
!    Department of Mechanical and Aerospace Engineering
!    Carleton University
!    1125 Colonel By Drive
!    Ottawa, ON, K1S 5B6
!    CANADA
!   
!    URL: http://www.mae.carleton.ca/Ron_Miller
!
!  -or-
!
!    Ellad B. Tadmor
!    Department of Aerospace Engineering and Mechanics
!    University of Minnesota
!    107 Akerman Hall, 110 Union St SE
!    Minneapolis, MN 55455
!    USA
!    
!    URL: http://www.aem.umn.edu/people/faculty/bio/Tadmor.shtml
!   
!-------------------------------------------------------------------------------

!****************************************************************************
!**
!**  MODULE mod_uservars   contains variables used by user routines
!**
!**                        Punch example
!**
!**  Variable Definitions:
!**  ---------------------
!**
!**  Integer Variables:
!**
!**    user_iparam          -  array of integers available for storage
!**                            of user-specified parameters. See also
!**                            user_rparam.
!**
!**  Real Variables:
!**
!**    user_rparam          -  array of reals available for storage
!**                            of user-specified parameters. See also
!**                            user_iparam.
!**
!****************************************************************************

      module mod_uservars
      use mod_global, only : dp

      save
      private
      public user_iparam, user_rparam

      integer user_iparam(10)
      real(kind=dp) user_rparam(10)

      end module mod_uservars


!**-----------------------------------------------------------------------------
!**
!**   user_mesh: Mesh generator routine for nanoindentation. 
!**              Creates an initial uniform mesh.
!**
!**   Calling format:
!**
!**   mesh,,nx,ny,contact
!**
!**   where  nx = number of x-divisions
!**          ny = number of y-divisions
!**          contact = indenter 1/2 width
!--
      subroutine user_mesh(id,x,ix,f,b,itx,key,input)
      use mod_uservars
      use mod_global,   only : dp,maxel,maxnp,ndf,ndm,nen,numel,numnp,
     &                         nxdm,SymmetricMesh,SymmetryDof,
     &                         SymmetryValue
      use mod_output,   only : erroroutput,nerror,output,string,error
      use mod_boundary, only : ncb,nce,NCEMAX,elist
      use mod_repatom,  only : protected
      use mod_qclib,    only : next,freein,qsortr
      use mod_mesh,     only : delaunay,PerturbMesh
      use mod_grain,    only : NumGrains,GetGrainNumVrts,GetGrainVertex,
     &                         NearestBSite
      implicit none

!-- Transferred Variables
      integer,           intent(inout)  :: id(ndf,maxnp),
     &                                     ix(nen,maxel),
     &                                     itx(3,maxel)
      real(kind=dp),     intent(inout)  :: x(nxdm,maxnp),
     &                                     f(ndf,maxnp),
     &                                     b(ndf,maxnp)
      character(len=4),  intent(in)     :: key
      character(len=80), intent(in)     :: input

!-- Local Variables
      real(kind=dp) xcind,ycind,contact
      real(kind=dp), allocatable :: vtmp(:)
      logical xflag,yflag
      integer nx,ny,kount,igr,i,j,k,idum,lower,upper,ivert,numvrts
      real(kind=dp) xlo,xhi,ylo,yhi,dx,dy,xx,yy,xtmp(3),xtmp1(2),
     & xmin,xmax,ymin,ymax,xc,yc,dum,vertex(ndm)
      real(kind=dp), parameter :: tol=1.e-3_dp

      ! Parse input
      lower = 0
      upper = next(lower,input)
      call freein(input,lower,upper,nx,dum,1)
      lower = upper
      upper = next(lower,input)
      call freein(input,lower,upper,ny,dum,1)
      lower = upper
      upper = next(lower,input)
      call freein(input,lower,upper,idum,contact,2)

      ! Echo input
      string=''; call output
      write(string,'(a)')'Generating uniform mesh for punch problem'
      call output
      string=''; call output
      write(string,'(a,i4)')'# x-elements        = ',nx
      call output
      write(string,'(a,i4)')'# y-elements        = ',ny
      call output
      write(string,'(a,f10.5)')'indenter 1/2 width  = ',contact
      call output

      ! This mesh generator is set up for one rectangular grain
      ! (make sure this is what we got)
      if (NumGrains().ne.1) then
         nerror = 2
         error(1) = 
     &    'More than one grain detected in subroutine user_mesh (punch)'
         error(2) = 'Routine desined to work only for one grain.'
         call erroroutput
         stop
      endif

      xlo=1.e30_dp
      xhi=-1.e30_dp
      ylo=1.e30_dp
      yhi=-1.e30_dp
      call GetGrainNumVrts(1,numvrts)
      do ivert=1,numvrts
         call GetGrainVertex(1,ivert,vertex)
         xx = vertex(1)
         yy = vertex(2)
         xlo = min(xx,xlo)
         xhi = max(xx,xhi)
         ylo = min(yy,ylo)
         yhi = max(yy,yhi)
      enddo

      dx = (xhi - xlo)/real(nx,dp)
      dy = (yhi - ylo)/real(ny,dp)

      ! Generate Nodes
      kount = 0
      do i = 0,nx
      do j = 0,ny
         xtmp1(1) = xlo + real(i,dp)*dx
         xtmp1(2) = ylo + real(j,dp)*dy
         call NearestBSite(xtmp1,.false.,xtmp,igr,idum,x,b,ix,itx)
         kount = kount + 1
         if (kount.gt.maxnp) then
            nerror = 1
            error(1) =
     &         'Number of nodes exceeds maxnp in subroutine user_mesh'
            call erroroutput
            stop
         endif
         x(1:3,kount) = xtmp(1:3)
      enddo
      enddo
      numnp = kount

      ! Identify model boundaries
      xmin = x(1,1)
      xmax = x(1,1)
      ymin = x(2,1)
      ymax = x(2,1)
      do i = 1,numnp
         xx = x(1,i)
         yy = x(2,i)
         if (xx.lt.xmin) xmin=xx
         if (xx.gt.xmax) xmax=xx
         if (yy.lt.ymin) ymin=yy
         if (yy.gt.ymax) ymax=yy
      enddo
      xc = (xmin+xmax)/2.0_dp
      yc = (ymin+ymax)/2.0_dp

      ! Remove duplicate nodes
      i = 1
      do while (i<numnp)
         j = i + 1
         do while (j<=numnp)
            if (abs( x(1,i) - x(1,j)) .lt. tol .and.
     &          abs( x(2,i) - x(2,j)) .lt. tol ) then
               x(:,j)=x(:,numnp)
               numnp = numnp - 1
            else
               j = j + 1
            endif
         enddo
         i = i + 1
      enddo

      ! Identify nodes lying on the boundaries and sort 
      ! in counterclockwise fashion for constrained Delaunay algorithm
      nce = 0
      allocate(vtmp(numnp))
      do i=1,numnp
         xx=x(1,i)
         yy=x(2,i)
         if (abs(xx-xmin).le.tol .or. abs(xx-xmax).le.tol .or.
     &       abs(yy-ymin).le.tol .or. abs(yy-ymax).le.tol) then
            nce=nce+1
            if (nce.gt.NCEMAX) then
               nerror = 1
               error(1) = 'nce exceeds NCEMAX in subroutine user_mesh'
               call erroroutput
               stop
            endif
            elist(1,nce)=i
            vtmp(i)=datan2(yy-yc,xx-xc)
         endif
      enddo
      call qsortr(nce,elist,vtmp,1.0_dp,NCEMAX,numnp)
      deallocate(vtmp)    

      ! Finish defining boundary
      elist(2,nce)=elist(1,1)
      ncb = nce
      do i=1,nce-1
         elist(2,i)=elist(1,i+1)
      enddo

      ! Triangulate
      SymmetricMesh = .true.   ! turn on symmetric meshing option
      SymmetryDof = 1
      SymmetryValue = 0.0_dp
      call PerturbMesh(x,.true.)
      call delaunay(x,ix,b,f,id,itx)
      call PerturbMesh(x,.false.)

      ! Set indenter center to middle of upper surface of model
      xcind = (xmin+xmax)/2
      ycind = ymax

      ! Boundary conditions (rigid base and symmetry BC's on sides. 
      ! Punch BC's are applied later in punchbc.f.
      do i = 1,numnp
         xx=x(1,i)
         yy=x(2,i)
         ! Protect initial nodes from deletion during coarsening
         protected(i) = .true.
         ! Rigid boundary at y=ymin
         if (abs(yy-ymin).le.tol) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
         endif
         ! Symmetry boundary at x=xmin
         if (abs(xx-xmin).le.tol) then
            id(1,i)=1
            id(3,i)=1
         endif
         ! Symmetry boundary at x=xmax
         if (abs(xx-xmax).le.tol) then
            id(1,i)=1
            id(3,i)=1
         endif
      enddo

      ! Report and exit
      string=''; call output
      write(string,'(a,i5)')'**numnp = ',numnp
      call output
      write(string,'(a,i5)')'**numel = ',numel
      call output

      ! Store punch location and size information for BC routine
      user_rparam(1) = xcind
      user_rparam(2) = ycind
      user_rparam(3) = contact

      return
      end


!**-----------------------------------------------------------------------------
!**
!**   user_bcon: Boundary Conditions routine for nanoindentation.
!**              Flat punch with perfect stick conditions.
!**
!**   Calling format:
!**
!**   bcon
!**
!**   Note: no parameters needed.
!--
      subroutine user_bcon(id,x,ix,f,jdiag,str,eps,q,b,dr,db,shp,
     &   xsj,key,input,flag)
      use mod_uservars
      use mod_global, only : dp,maxel,maxneq,maxnp,ndf,nen,nq,nstr,
     &                       numnp,nxdm
      use mod_output, only : output,string
      use mod_pload,  only : GetLoadPropFact,GetLoadOldPropFact,
     &                       GetLoadTime,GetLoadTimeStep
      implicit  none
 
!-- Transferred Variables
      real(kind=dp),     intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel)
      integer,           intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp)
      character(len=80), intent(in)    :: input
      character(len=4),  intent(in)    :: key
      logical,           intent(inout) :: flag

!-- Local Variables
      real(kind=dp), parameter :: tole=1.e-7_dp
      real(kind=dp) xcind,ycind,contact
      real(kind=dp) xx,yy,propfact
      integer i

      ! Extract punch information stored in user param array 
      ! by meshing routine
      xcind = user_rparam(1)
      ycind = user_rparam(2)
      contact = user_rparam(3)

      ! Constrain displacements of all nodes under indenter to
      ! appropriate values. List of nodes in contact with the indenter
      ! may change due to adaption - so need to look for them.
      propfact = GetLoadPropFact()
      do i=1,numnp
         xx=x(1,i)
         yy=x(2,i)
         if (abs(yy-ycind).le.tole.and.abs(xx-xcind).le.contact) then
            id(1,i)=1
            id(2,i)=1
            id(3,i)=1
            f(2,i)=-1.0_dp
            b(2,i)=f(2,i)*propfact
         endif
      enddo

      ! output
      write(string,'(a)')'* Boundary conditions applied:'
      call output
      write(string,'(2x,a,f5.2)')
     &             'nodes under indenter displaced down by ',propfact
      call output

      return
      end


!**---------------------------------------------------------------
!**   user_pdel : compute the total external load for the p-delta 
!**               curve and for convergence checks.
!**
!--
      subroutine user_pdel(id,x,ix,f,jdiag,str,eps,q,b,dr,db,shp,
     &   xsj,key,input,force)
      use mod_uservars
      use mod_global, only : dp,maxel,maxneq,maxnp,ndf,nen,nq,nstr,
     &                       numnp,nxdm
      implicit  none

!-- Transferred Variables
      real(kind=dp),     intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel)
      integer,           intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp)
      character(len=80), intent(in)    :: input
      character(len=4),  intent(in)    :: key
      real(kind=dp),     intent(out)   :: force

!-- Local Variables
      integer i,j
      real(kind=dp) ff
 
      force = 0.0_dp
      do j = 1,numnp
         do i = 1,ndf
            if (id(i,j).eq.1) then
               ff = 0.0_dp
               if (f(i,j).lt.-1.0e-9_dp) ff=-1.0_dp
               if (f(i,j).gt.1.0e-9_dp) ff=1.0_dp
               force = force - ff*dr(i,j)
            endif
         enddo
      enddo

      return
      end


!**---------------------------------------------------------------
!**   user_potential : user-defined external potential 
!**
!**
!**
      subroutine user_potential(id,x,ix,f,jdiag,str,eps,q,b,dr,shp,xsj,
     &                          fls,flt,flw)
      use mod_uservars
      use mod_global, only : dp,maxneq,maxnp,maxel,ndf,nen,nq,nstr,nxdm,
     &                       user_energy
      use mod_pload,  only : GetLoadPropFact,GetLoadOldPropFact,
     &                       GetLoadTime,GetLoadTimeStep
      use mod_stiff,  only : tang,istiff
      implicit  none

!-- Transferred Variables
      real(kind=dp), intent(inout) :: b(ndf,maxnp),str(nstr,maxel),
     &                                eps(nstr,maxel),q(nq,maxel),
     &                                x(nxdm,maxnp),f(ndf,maxnp),
     &                                dr(ndf,maxnp),shp(3,nen,maxel),
     &                                xsj(maxel)
      integer,       intent(inout) :: jdiag(maxneq),ix(nen,maxel),
     &                                id(ndf,maxnp)
      logical,       intent(in)    :: fls,flt,flw

!-- Local Variables

      user_energy = 0.0_dp

      return
      end

!--------------------------------------------------------------------
!
!     user_plot : User-specified output plot
!
!     Calling format:
!
!     plot,user,filepref,[index],[umag],[scale],[append]
!
!
!
      subroutine user_plot(id,x,ix,itx,f,jdiag,str,eps,q,b,dr,db,shp,
     &                     xsj,index,umag,scale,logic,input,upper)
      use mod_uservars
      use mod_global, only : dp,maxnp,maxel,maxneq,ndf,nen,nq,nstr,nxdm
      implicit  none

!-- Transferred Variables
      real(kind=dp),     intent(in)    :: b(ndf,maxnp),str(nstr,maxel),
     &                                    eps(nstr,maxel),q(nq,maxel),
     &                                    x(nxdm,maxnp),f(ndf,maxnp),
     &                                    dr(ndf,maxnp),db(maxneq),
     &                                    shp(3,nen,maxel),xsj(maxel),
     &                                    umag,scale
      integer,           intent(in)    :: jdiag(maxneq),ix(nen,maxel),
     &                                    id(ndf,maxnp),itx(3,maxel),
     &                                    index,logic
      integer,           intent(inout) :: upper
      character(len=80), intent(in)    :: input

      return
      end
